# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Factory method for creating new action needs."""

from __future__ import absolute_import, print_function

from functools import partial

from flask_principal import ActionNeed

from .permissions import ParameterizedActionNeed


def action_factory(name, parameter=False):
    """Factory method for creating new actions (w/wo parameters).

    :param name: Name of the action (prefix with your module name).
    :param parameter: Determines if action should take parameters or not.
        Default is ``False``.
    """
    if parameter:
        return partial(ParameterizedActionNeed, name)
    else:
        return ActionNeed(name)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for common role based access control."""

from __future__ import absolute_import, print_function

import pkg_resources
import six
from flask_principal import identity_loaded
from werkzeug.utils import cached_property, import_string

from . import config
from .loaders import load_permissions_on_identity_loaded


class _AccessState(object):
    """Access state storing registered actions."""

    def __init__(self, app, entry_point_actions=None,
                 entry_point_system_roles=None, cache=None):
        """Initialize state.

        :param app: The Flask application.
        :param entry_point_actions: The entrypoint for actions extensions.
            (Default: ``None``)
        :param entry_point_system_roles: The entrypoint for system roles
            extensions. (Default: ``None``)
        :param cache: The cache system. (Default: ``None``)
        """
        self.app = app
        self.actions = {}
        self.system_roles = {}
        self._cache = cache
        if entry_point_actions:
            self.load_entry_point_actions(entry_point_actions)
        if entry_point_system_roles:
            self.load_entry_point_system_roles(entry_point_system_roles)

    @cached_property
    def cache(self):
        """Return a cache instance."""
        cache = self._cache or self.app.config.get('ACCESS_CACHE')
        return import_string(cache) if isinstance(cache, six.string_types) \
            else cache

    def set_action_cache(self, action_key, data):
        """Store action needs and excludes.

        .. note:: The action is saved only if a cache system is defined.

        :param action_key: The unique action name.
        :param data: The action to be saved.
        """
        if self.cache:
            self.cache.set(
                self.app.config['ACCESS_ACTION_CACHE_PREFIX'] +
                action_key, data
            )

    def get_action_cache(self, action_key):
        """Get action needs and excludes from cache.

        .. note:: It returns the action if a cache system is defined.

        :param action_key: The unique action name.
        :returns: The action stored in cache or ``None``.
        """
        data = None
        if self.cache:
            data = self.cache.get(
                self.app.config['ACCESS_ACTION_CACHE_PREFIX'] +
                action_key
            )
        return data

    def delete_action_cache(self, action_key):
        """Delete action needs and excludes from cache.

        .. note:: It returns the action if a cache system is defined.

        :param action_key: The unique action name.
        """
        if self.cache:
            self.cache.delete(
                self.app.config['ACCESS_ACTION_CACHE_PREFIX'] +
                action_key
            )

    def register_action(self, action):
        """Register an action to be showed in the actions list.

        .. note:: A action can't be registered two times. If it happens, then
        an assert exception will be raised.

        :param action: The action to be registered.
        """
        assert action.value not in self.actions
        self.actions[action.value] = action

    def load_entry_point_actions(self, entry_point_group):
        """Load actions from an entry point group.

        :param entry_point_group: The entrypoint for extensions.
        """
        for ep in pkg_resources.iter_entry_points(group=entry_point_group):
            self.register_action(ep.load())

    def register_system_role(self, system_role):
        """Register a system role.

        .. note:: A system role can't be registered two times. If it happens,
        then an assert exception will be raised.

        :param system_role: The system role to be registered.
        """
        assert system_role.value not in self.system_roles
        self.system_roles[system_role.value] = system_role

    def load_entry_point_system_roles(self, entry_point_group):
        """Load system roles from an entry point group.

        :param entry_point_group: The entrypoint for extensions.
        """
        for ep in pkg_resources.iter_entry_points(group=entry_point_group):
            self.register_system_role(ep.load())


class InvenioAccess(object):
    """Invenio Access extension."""

    def __init__(self, app=None, **kwargs):
        """Extension initialization.

        :param app: The Flask application. (Default: ``None``)
        """
        if app:
            self._state = self.init_app(app, **kwargs)

    def init_app(self, app, entry_point_actions='invenio_access.actions',
                 entry_point_system_roles='invenio_access.system_roles',
                 **kwargs):
        """Flask application initialization.

        :param app: The Flask application.
        :param entry_point_actions: The entrypoint for actions extensions.
            (Default: ``'invenio_access.actions'``)
        :param entry_point_system_roles: The entrypoint for  system roles
            extensions. (Default: ``'invenio_access.system_roles'``)
        :param cache: The cache system. (Default: ``None``)
        """
        self.init_config(app)
        state = _AccessState(
            app, entry_point_actions=entry_point_actions,
            entry_point_system_roles=entry_point_system_roles,
            cache=kwargs.get('cache'))
        app.extensions['invenio-access'] = state

        if app.config.get('ACCESS_LOAD_SYSTEM_ROLE_NEEDS', True):
            identity_loaded.connect_via(app)(
                load_permissions_on_identity_loaded
            )

        return state

    def init_config(self, app):
        """Initialize configuration.

        :param app: The Flask application.
        """
        for k in dir(config):
            if k.startswith('ACCESS_'):
                app.config.setdefault(k, getattr(config, k))

    def __getattr__(self, name):
        """Proxy to state object."""
        return getattr(self._state, name, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Admin views for managing access to actions."""

from flask import current_app
from flask_admin.contrib.sqla import ModelView
from werkzeug.local import LocalProxy
from wtforms import SelectField

from .models import ActionRoles, ActionSystemRoles, ActionUsers
from .proxies import current_access

_datastore = LocalProxy(lambda: current_app.extensions['security'].datastore)


def _(x):
    """Identity."""
    return x


class ActionUsersView(ModelView):
    """View for managing access to actions by users."""

    can_view_details = True

    list_all = ('user_id', 'user.email', 'action', 'argument', 'exclude')

    column_list = list_all

    column_default_sort = ('user_id', True)

    column_labels = {
        'user.email': _("Email"),
        'user_id': _("User ID"),
        'exclude': _("Deny"),
    }

    column_filters = list_all

    form_columns = ('user', 'action', 'argument', 'exclude')

    form_args = dict(
        action=dict(
            choices=LocalProxy(lambda: [
                (action, action) for action in current_access.actions.keys()
            ])
        )
    )
    form_overrides = dict(
        action=SelectField,
    )


class ActionRolesView(ModelView):
    """View for managing access to actions by users with given role."""

    can_view_details = True

    list_all = ('role.name', 'action', 'argument', 'exclude')

    column_list = list_all

    column_filters = \
        columns_sortable_list = \
        columns_searchable_list = \
        list_all

    column_display_all_relations = True

    column_labels = {
        'role.name': _("Role Name"),
        'exclude': _("Deny"),
    }

    form_columns = ('role', 'action', 'argument', 'exclude')

    form_args = dict(
        action=dict(
            choices=LocalProxy(lambda: [
                (action, action) for action in current_access.actions.keys()
            ])
        )
    )

    form_overrides = dict(
        action=SelectField,
    )


class ActionSystemRolesView(ModelView):
    """View for managing access to actions by users with system roles."""

    can_view_details = True

    list_all = ('role_name', 'action', 'argument', 'exclude')

    column_list = list_all

    column_filters = \
        columns_sortable_list = \
        columns_searchable_list = \
        list_all

    column_display_all_relations = True

    column_labels = {
        'role_name': _("System Role"),
        'exclude': _("Deny"),
    }

    form_args = dict(
        action=dict(
            choices=LocalProxy(lambda: [
                (action, action) for action in current_access.actions.keys()
            ])
        ),
        role_name=dict(
            choices=LocalProxy(lambda: [
                (action, action) for action
                in current_access.system_roles.keys()
            ])
        )
    )

    form_columns = ('role_name', 'action', 'argument', 'exclude')
    form_overrides = dict(
        action=SelectField,
        role_name=SelectField,
    )


action_roles_adminview = {
    'model': ActionRoles,
    'modelview': ActionRolesView,
    'category': _('User Management'),
    'name': _('Access: Roles')
}

action_users_adminview = {
    'model': ActionUsers,
    'modelview': ActionUsersView,
    'category': _('User Management'),
    'name': _('Access: Users')
}

action_system_roles_adminview = {
    'model': ActionSystemRoles,
    'modelview': ActionSystemRolesView,
    'category': _('User Management'),
    'name': _('Access: System Roles')
}

__all__ = ('action_users_adminview', 'action_roles_adminview',
           'action_system_roles_adminview')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Pytest configuration."""

from __future__ import absolute_import, print_function

import os
import warnings

import pytest
from flask import Flask
from flask.cli import ScriptInfo
from flask_babelex import Babel
from flask_mail import Mail
from flask_principal import ActionNeed, identity_loaded
from invenio_accounts import InvenioAccounts
from invenio_db import InvenioDB, db
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import DropConstraint, DropSequence, DropTable

from invenio_access import InvenioAccess
from invenio_access.loaders import load_permissions_on_identity_loaded
from invenio_access.permissions import ParameterizedActionNeed, Permission


@compiles(DropTable, 'postgresql')
def _compile_drop_table(element, compiler, **kwargs):
    return compiler.visit_drop_table(element) + ' CASCADE'


@compiles(DropConstraint, 'postgresql')
def _compile_drop_constraint(element, compiler, **kwargs):
    return compiler.visit_drop_constraint(element) + ' CASCADE'


@compiles(DropSequence, 'postgresql')
def _compile_drop_sequence(element, compiler, **kwargs):
    return compiler.visit_drop_sequence(element) + ' CASCADE'


@pytest.fixture()
def base_app():
    """Flask base application fixture."""
    app_ = Flask('testapp')
    app_.config.update(
        ACCOUNTS_USE_CELERY=False,
        SECRET_KEY='CHANGE_ME',
        SECURITY_PASSWORD_SALT='CHANGE_ME_ALSO',
        SQLALCHEMY_DATABASE_URI=os.environ.get(
            'SQLALCHEMY_DATABASE_URI', 'sqlite:///test.db'),
        SQLALCHEMY_TRACK_MODIFICATIONS=False,
        TESTING=True,
    )
    Babel(app_)
    Mail(app_)
    InvenioDB(app_)
    InvenioAccounts(app_)
    return app_


@pytest.fixture()
def app(base_app, request):
    """Flask application fixture."""
    with base_app.app_context():
        db.create_all()

    def teardown():

        with base_app.app_context():
            db.drop_all()
            identity_loaded.disconnect(load_permissions_on_identity_loaded)

    request.addfinalizer(teardown)
    return base_app


@pytest.fixture()
def access_app(app):
    """Init InvenioAccess app and return a request context."""
    InvenioAccess(app)
    with app.test_request_context():
        yield app


@pytest.fixture()
def script_info(app):
    """Get ScriptInfo object for testing CLI."""
    action_open = ActionNeed('open')
    action_edit = ParameterizedActionNeed('edit', None)

    ext = InvenioAccess(app)
    ext.register_action(action_open)
    ext.register_action(action_edit)
    return ScriptInfo(create_app=lambda info: app)


@pytest.fixture()
def script_info_cli_list(app):
    """Get ScriptInfo object for testing CLI list command."""
    action_open = ActionNeed('open')
    action_edit = ParameterizedActionNeed('edit', None)
    ext = InvenioAccess(app)
    ext.register_action(action_open)
    ext.register_action(action_edit)

    return ScriptInfo(create_app=lambda info: app)


@pytest.fixture()
def dynamic_permission():
    """Dynamic permission fixture."""
    def _get_dynamic_permission(*args):
        """Get dynamic permission."""
        return DynamicPermission(*args)
    return _get_dynamic_permission


class DynamicPermission(Permission):
    """Represents set of required needs.

    Works like :py:class:`~.Permission` except that any action not
    allowed/restricted to any users, roles or system roles are allowed by
    default instead of restricted.

    .. deprecated:: 1.0.0

        DynamicPermission is deprecated in favor of :py:class:`~.Permission`.

    .. warning::

        This class is going to be removed in a future version.

        The class works significantly different from normal permission class in
        that if ``ActionNeed`` or :py:data:`~.ParameterizedActionNeed` is not
        allowed or restricted to any user or role then it is **ALLOWED** to
        anybody.
    """

    allow_by_default = True

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super(DynamicPermission, self).__init__(*args, **kwargs)
        warnings.warn("DynamicPermission is scheduled for removal.",
                      DeprecationWarning)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Helper proxy to the state object."""

from __future__ import absolute_import, print_function

from flask import current_app
from werkzeug.local import LocalProxy

current_access = LocalProxy(
    lambda: current_app.extensions['invenio-access']
)
"""Helper proxy to access state object."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import pytest
from flask import Flask
from flask_babelex import Babel
from flask_mail import Mail
from flask_principal import ActionNeed
from invenio_accounts import InvenioAccounts
from invenio_db import InvenioDB, db
from invenio_db.utils import drop_alembic_version_table
from mock import patch
from pkg_resources import EntryPoint

from invenio_access import InvenioAccess, current_access
from invenio_access.permissions import SystemRoleNeed


class MockEntryPointAction(EntryPoint):
    """Mocking of entrypoint."""

    def load(self):
        """Mock load entry point."""
        return ActionNeed(self.name)


class MockEntryPointSystemRole(EntryPoint):
    """Mocking of entrypoint."""

    def load(self):
        """Mock load entry point."""
        return SystemRoleNeed(self.name)


def _mock_entry_points(group=None):
    """Mocking funtion of entrypoints."""
    data = {
        'invenio_access.actions': [
            MockEntryPointAction('open', 'demo.actions'),
            MockEntryPointAction('close', 'demo.actions')
        ],
        'invenio_access.system_roles': [
            MockEntryPointSystemRole('any_user', 'demo.specrole'),
            MockEntryPointSystemRole('authenticated_user', 'demo.specrole')
        ],
    }
    names = data.keys() if group is None else [group]
    for key in names:
        for entry_point in data[key]:
            yield entry_point


def test_version():
    """Test version import."""
    from invenio_access import __version__
    assert __version__


def test_init(base_app):
    """Test extension initialization."""
    app = base_app
    ext = InvenioAccess(app)
    assert 'invenio-access' in app.extensions


def test_init_app(base_app):
    """Test extension initialization."""
    app = base_app
    ext = InvenioAccess()
    assert 'invenio-access' not in app.extensions
    ext.init_app(app)
    assert 'invenio-access' in app.extensions


def test_actions(base_app):
    """Test if the actions are registered properly."""
    InvenioAccess(base_app, entry_point_actions=None)
    with base_app.app_context():
        current_access.register_action(ActionNeed('action_a'))
        assert len(current_access.actions) == 1
        current_access.register_action(ActionNeed('action_b'))
        assert len(current_access.actions) == 2


def test_system_roles(base_app):
    """Test if the system roles are registered properly."""
    InvenioAccess(base_app, entry_point_system_roles=None)
    with base_app.app_context():
        current_access.register_system_role(SystemRoleNeed('spn_a'))
        assert len(current_access.system_roles) == 1
        current_access.register_system_role(SystemRoleNeed('spn_b'))
        assert len(current_access.system_roles) == 2


@patch('pkg_resources.iter_entry_points', _mock_entry_points)
def test_entrypoints():
    """Test if the entrypoints are registering actions and roles properly."""
    app = Flask('testapp')
    ext = InvenioAccess(app)
    assert len(ext.actions) == 2
    assert ActionNeed('open') in ext.actions.values()
    assert ActionNeed('close') in ext.actions.values()
    assert len(ext.system_roles) == 2
    assert SystemRoleNeed('any_user') in ext.system_roles.values()
    assert SystemRoleNeed('authenticated_user') in ext.system_roles.values()


def test_alembic(app):
    """Test alembic recipes."""
    ext = app.extensions['invenio-db']

    with app.app_context():
        if db.engine.name == 'sqlite':
            raise pytest.skip('Upgrades are not supported on SQLite.')

        assert not ext.alembic.compare_metadata()
        db.drop_all()
        drop_alembic_version_table()
        ext.alembic.upgrade()

        assert not ext.alembic.compare_metadata()
        ext.alembic.downgrade(target='96e796392533')
        ext.alembic.upgrade()

        assert not ext.alembic.compare_metadata()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add ActionSystemRoles."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '04480be1593e'
down_revision = 'dbe499bfda43'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        'access_actionssystemroles',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('action', sa.String(length=80), nullable=True),
        sa.Column('exclude', sa.Boolean(name='exclude'), server_default='0',
                  nullable=False),
        sa.Column('argument', sa.String(length=255), nullable=True),
        sa.Column('role_name', sa.String(length=40), nullable=False),
        sa.PrimaryKeyConstraint('id',
                                name=op.f('pk_access_actionssystemroles')),
        sa.UniqueConstraint('action', 'exclude', 'argument', 'role_name',
                            name='access_actionssystemroles_unique')
    )
    op.create_index(op.f('ix_access_actionssystemroles_action'),
                    'access_actionssystemroles', ['action'], unique=False)
    op.create_index(op.f('ix_access_actionssystemroles_argument'),
                    'access_actionssystemroles', ['argument'], unique=False)
    op.create_index(op.f('ix_access_actionssystemroles_role_name'),
                    'access_actionssystemroles', ['role_name'], unique=False)


def downgrade():
    """Downgrade database."""
    op.drop_index(op.f('ix_access_actionssystemroles_role_name'),
                  table_name='access_actionssystemroles')
    op.drop_index(op.f('ix_access_actionssystemroles_argument'),
                  table_name='access_actionssystemroles')
    op.drop_index(op.f('ix_access_actionssystemroles_action'),
                  table_name='access_actionssystemroles')
    op.drop_table('access_actionssystemroles')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Default values for access configuration.

.. note::

    By default no caching is enabled. For production instances it is **highly
    advisable** to enable caching as the permission checking is very query
    intensive on the database.
"""

ACCESS_CACHE = None
"""A cache instance or an importable string pointing to the cache instance."""

ACCESS_ACTION_CACHE_PREFIX = 'Permission::action::'
"""Prefix for actions cached when used in dynamic permissions."""

ACCESS_LOAD_SYSTEM_ROLE_NEEDS = True
"""Enables the loading of system role needs when users' identity change."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Permission and action needs for Invenio."""

from collections import namedtuple
from functools import partial
from itertools import chain

from flask_principal import ActionNeed, Need
from flask_principal import Permission as _Permission

from .models import ActionRoles, ActionSystemRoles, ActionUsers, \
    get_action_cache_key
from .proxies import current_access

_Need = namedtuple("Need", ["method", "value", "argument"])

ParameterizedActionNeed = partial(_Need, "action")

ParameterizedActionNeed.__doc__ = \
    """A need having the method preset to `"action"` and a parameter.

    If it is called with `argument=None` then this need is equivalent
    to ``ActionNeed``.
    """

SystemRoleNeed = partial(Need, "system_role")
SystemRoleNeed.__doc__ = \
    """A need with the method preset to `"system_role"`."""

#
# Need instances
#
superuser_access = ActionNeed("superuser-access")
"""Superuser access aciton which allow access to everything."""

any_user = SystemRoleNeed("any_user")
"""Any user system role.

This role is used to assign all possible users (authenticated and guests)
to an action.
"""

authenticated_user = SystemRoleNeed("authenticated_user")
"""Authenticated user system role.

This role is used to assign all authenticated users to an action.
"""


class _P(namedtuple("Permission", ["needs", "excludes"])):
    """Helper for simple permission updates."""

    def update(self, permission):
        """In-place update of permissions."""
        self.needs.update(permission.needs)
        self.excludes.update(permission.excludes)


class Permission(_Permission):
    """Represents a set of required needs.

    Extends Flask-Principal's :py:class:`flask_principal.Permission` with
    support for loading action grants from the database including caching
    support.

    Essentially the class works as a translation layer that expands action
    needs into a list of user/roles needs. For instance, take the following
    permission:

    .. code-block:: python

        Permission(ActionNeed('my-action'))

    Once the permission is checked with an identity, the class will fetch a
    list of all users and roles that have been granted/denied access to the
    action, and expand the permission into something similar to (depending
    on the state of the database):

    .. code-block:: python

        Permission(UserNeed('1'), RoleNeed('admin'))

    The expansion is cached until the action is modified (e.g. a user is
    granted access to the action). The alternative approach to expanding the
    action need like this class is doing, would be to load the list of allowed
    actions for a user on login and cache the result. However retrieving all
    allowed actions for a  user could results in very large lists, where as
    caching allowed users/roles for an action would usually yield smaller lists
    (especially if roles are used).
    """

    allow_by_default = False
    """If enabled, all permissions are granted when they are not assigned to
    anybody. Disabled by default.
    """

    def __init__(self, *needs):
        r"""Initialize permission.

        :param \*needs: The needs for this permission.
        """
        self._permissions = None
        self.explicit_needs = set(needs)
        self.explicit_needs.add(superuser_access)
        self.explicit_excludes = set()

    @staticmethod
    def _cache_key(action_need):
        """Helper method to generate cache key."""
        return get_action_cache_key(
            action_need.value,
            action_need.argument if hasattr(action_need, "argument") else None,
        )

    @staticmethod
    def _split_actionsneeds(needs):
        """Split needs into sets of ActionNeed and any other *Need."""
        action_needs, other_needs = set(), set()
        for need in needs:
            if need.method == "action":
                action_needs.add(need)
            else:
                other_needs.add(need)
        return action_needs, other_needs

    def _load_permissions(self):
        """Load permissions for all needs, expanding actions."""
        result = _P(needs=set(), excludes=set())

        # split ActionNeeds and any other Need in separates Sets
        action_needs, explicit_needs = self._split_actionsneeds(
            self.explicit_needs)
        action_excludes, explicit_excludes = self._split_actionsneeds(
            self.explicit_excludes)

        # add all explicit needs/excludes to the result permissions
        result.needs.update(explicit_needs)
        result.excludes.update(explicit_excludes)

        # expand all ActionNeeds to get all needs/excludes and add them to the
        # result permissions
        for need in action_needs | action_excludes:
            result.update(self._expand_action(need))

        # "allow_by_default = False" means that when needs are empty,
        # then it should deny access.
        # By default, `flask_principal.Permission.allows` will allow access
        # if needs are empty!
        needs_empty = len(result.needs) == 0
        deny_access_when_empty_needs = not self.allow_by_default
        if needs_empty and deny_access_when_empty_needs:
            # Add at least one dummy need so that it will always deny access
            result.needs.update(action_needs)

        self._permissions = result

    def _expand_action(self, explicit_action):
        """Expand action to user/roles needs and excludes."""
        action = current_access.get_action_cache(
            self._cache_key(explicit_action)
        )
        if action is None:
            action = _P(needs=set(), excludes=set())

            actionsusers = ActionUsers.query_by_action(explicit_action).all()

            actionsroles = (
                ActionRoles.query_by_action(explicit_action)
                .join(ActionRoles.role)
                .all()
            )

            actionssystem = ActionSystemRoles.query_by_action(
                explicit_action
            ).all()

            for db_action in chain(actionsusers, actionsroles, actionssystem):
                if db_action.exclude:
                    action.excludes.add(db_action.need)
                else:
                    action.needs.add(db_action.need)

            current_access.set_action_cache(
                self._cache_key(explicit_action), action
            )
        return action

    @property
    def needs(self):
        """Return allowed permissions from database.

        :returns: A list of need instances.
        """
        self._load_permissions()
        return self._permissions.needs

    @property
    def excludes(self):
        """Return denied permissions from database.

        :returns: A list of need instances.
        """
        self._load_permissions()
        return self._permissions.excludes



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import time

from flask_principal import ActionNeed, Need, RoleNeed, UserNeed
from invenio_accounts.models import Role, User
from invenio_db import db
from werkzeug.contrib.cache import RedisCache, SimpleCache

from invenio_access import InvenioAccess, current_access
from invenio_access.models import ActionRoles, ActionSystemRoles, ActionUsers
from invenio_access.permissions import ParameterizedActionNeed, SystemRoleNeed


class FakeIdentity(object):
    """Fake class to test DynamicPermission."""

    def __init__(self, *provides):
        self.provides = provides


def test_invenio_access_permission_cache(app, dynamic_permission):
    """Caching the user using memory caching."""
    cache = SimpleCache()
    InvenioAccess(app, cache=cache)
    with app.test_request_context():
        user_can_all = User(email='all@inveniosoftware.org')
        user_can_open = User(email='open@inveniosoftware.org')
        user_can_open_1 = User(email='open1@inveniosoftware.org')

        db.session.add(user_can_all)
        db.session.add(user_can_open)
        db.session.add(user_can_open_1)

        db.session.add(ActionUsers(action='open', user=user_can_all))

        db.session.flush()

        permission_open = dynamic_permission(ActionNeed('open'))

        identity_open = FakeIdentity(UserNeed(user_can_open.id))

        assert not permission_open.allows(identity_open)
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1)]),
            set([])
        )

        db.session.add(ActionUsers(action='open', user=user_can_open))
        db.session.flush()

        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_open)
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2)]),
            set([])
        )

        db.session.add(ActionUsers(action='open', argument=1,
                                   user=user_can_open_1))
        db.session.flush()

        identity_open_1 = FakeIdentity(UserNeed(user_can_open_1.id))
        permission_open_1 = dynamic_permission(
            ParameterizedActionNeed('open', '1'))
        assert not permission_open.allows(identity_open_1)
        assert permission_open_1.allows(identity_open_1)
        assert current_access.get_action_cache('open::1') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2),
                 Need(method='id', value=3)]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2)]),
            set([])
        )


def test_invenio_access_permission_cache_redis(app, dynamic_permission):
    """Caching the user using redis."""
    cache = RedisCache()
    InvenioAccess(app, cache=cache)
    with app.test_request_context():
        user_can_all = User(email='all@inveniosoftware.org')
        user_can_open = User(email='open@inveniosoftware.org')
        user_can_open_1 = User(email='open1@inveniosoftware.org')

        db.session.add(user_can_all)
        db.session.add(user_can_open)
        db.session.add(user_can_open_1)

        db.session.add(ActionUsers(action='open', user=user_can_all))

        db.session.flush()

        identity_open = FakeIdentity(UserNeed(user_can_open.id))

        permission_open = dynamic_permission(ActionNeed('open'))
        assert not permission_open.allows(identity_open)
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1)]),
            set([])
        )

        db.session.add(ActionUsers(action='open', user=user_can_open))
        db.session.flush()

        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_open)
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2)]),
            set([])
        )

        db.session.add(ActionUsers(action='open', argument=1,
                                   user=user_can_open_1))

        db.session.flush()

        identity_open_1 = FakeIdentity(UserNeed(user_can_open_1.id))
        permission_open_1 = dynamic_permission(
            ParameterizedActionNeed('open', '1'))
        assert not permission_open.allows(identity_open_1)
        assert permission_open_1.allows(identity_open_1)
        assert current_access.get_action_cache('open::1') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2),
                 Need(method='id', value=3)]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2)]),
            set([])
        )


def test_intenio_access_cache_performance(app, dynamic_permission):
    """Performance test simulating 1000 users."""
    InvenioAccess(app, cache=None)
    with app.test_request_context():
        # CDS has (2015-11-19) 74 actions with 414 possible arguments with
        # 49259 users and 307 roles. In this test we are going to divide
        # into 50 and use the next prime number.
        users_number = 991
        actions_users_number = 11
        actions_roles_number = 7

        roles = []
        actions = []
        for i in range(actions_roles_number):
            role = Role(name='role{0}'.format(i))
            roles.append(role)
            db.session.add(role)
            db.session.flush()

            action_role = ActionRoles(
                action='action{0}'.format(
                    str(i % actions_roles_number)),
                role=role)
            actions.append(action_role)
            db.session.add(action_role)
            db.session.flush()

        users = []
        for i in range(users_number):
            user = User(
                email='invenio{0}@inveniosoftware.org'.format(str(i)),
                roles=[roles[i % actions_roles_number]])
            users.append(user)
            db.session.add(user)
            db.session.flush()

            action_user = ActionUsers(action='action{0}'.format(
                str((i % actions_users_number) + actions_roles_number)),
                user=user)
            actions.append(action_user)
            db.session.add(action_user)
            db.session.flush()

        def test_permissions():
            """Iterates over all users checking its permissions."""
            for i in range(users_number):
                identity = FakeIdentity(UserNeed(users[i].id))

                # Allowed permission
                permission_allowed_both = dynamic_permission(
                    ActionNeed('action{0}'.format(
                        (i % actions_users_number) +
                        actions_roles_number)),
                    ActionNeed('action{0}'.format(i % actions_roles_number))
                )
                assert permission_allowed_both.allows(identity)

                # Not allowed action user
                permission_not_allowed_user = dynamic_permission(
                    ActionNeed('action{0}'.format(
                        (i + 1) % actions_users_number +
                        actions_roles_number))
                )
                assert not permission_not_allowed_user.allows(identity)

                # Not allowed action role
                permission_not_allowed_role = dynamic_permission(
                    ActionNeed('action{0}'.format(
                        (i + 1) % actions_roles_number))
                )
                assert not permission_not_allowed_role.allows(identity)

        app.extensions['invenio-access'].cache = None
        start_time_wo_cache = time.time()
        test_permissions()
        end_time_wo_cache = time.time()
        time_wo_cache = end_time_wo_cache - start_time_wo_cache

        app.extensions['invenio-access'].cache = SimpleCache()
        start_time_w_cache = time.time()
        test_permissions()
        end_time_w_cache = time.time()
        time_w_cache = end_time_w_cache - start_time_w_cache

        assert time_wo_cache / time_w_cache > 10


def test_invenio_access_permission_cache_users_updates(
    app,
    dynamic_permission
):
    """Testing ActionUsers cache with inserts/updates/deletes."""
    cache = SimpleCache()
    InvenioAccess(app, cache=cache)
    with app.test_request_context():
        # Creation of some data to test.
        user_1 = User(email='user_1@inveniosoftware.org')
        user_2 = User(email='user_2@inveniosoftware.org')
        user_3 = User(email='user_3@inveniosoftware.org')
        user_4 = User(email='user_4@inveniosoftware.org')
        user_5 = User(email='user_5@inveniosoftware.org')
        user_6 = User(email='user_6@inveniosoftware.org')

        db.session.add(user_1)
        db.session.add(user_2)
        db.session.add(user_3)
        db.session.add(user_4)
        db.session.add(user_5)
        db.session.add(user_6)

        db.session.add(ActionUsers(action='open', user=user_1))
        db.session.add(ActionUsers(action='write', user=user_4))

        db.session.flush()

        # Creation identities to test.
        identity_user_1 = FakeIdentity(UserNeed(user_1.id))
        identity_user_2 = FakeIdentity(UserNeed(user_2.id))
        identity_user_3 = FakeIdentity(UserNeed(user_3.id))
        identity_user_4 = FakeIdentity(UserNeed(user_4.id))
        identity_user_5 = FakeIdentity(UserNeed(user_5.id))
        identity_user_6 = FakeIdentity(UserNeed(user_6.id))

        # Test if user 1 can open. In this case, the cache should store only
        # this object.
        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_user_1)
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1)]),
            set([])
        )

        # Test if user 4 can write. In this case, the cache should have this
        # new object and the previous one (Open is allowed to user_1)
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_user_4)
        assert current_access.get_action_cache('write') == (
            set([Need(method='id', value=4)]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1)]),
            set([])
        )

        # If we add a new user to the action open, the open action in cache
        # should be removed but it should still containing the write entry.
        db.session.add(ActionUsers(action='open', user=user_2))
        db.session.flush()
        assert current_access.get_action_cache('open') is None
        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_user_2)
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2)]),
            set([])
        )
        assert current_access.get_action_cache('write') == (
            set([Need(method='id', value=4)]),
            set([])
        )

        # Test if the new user is added to the action 'open'
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_user_4)
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2)]),
            set([])
        )
        assert current_access.get_action_cache('write') == (
            set([Need(method='id', value=4)]),
            set([])
        )

        # If we update an action swapping a user, the cache containing the
        # action, should be removed.
        user_4_action_write = ActionUsers.query.filter(
            ActionUsers.action == 'write' and
            ActionUsers.user == user_4).first()
        user_4_action_write.user = user_3
        db.session.flush()
        assert current_access.get_action_cache('write') is None
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2)]),
            set([])
        )

        # Test if the user_3 can now write.
        permission_write = dynamic_permission(ActionNeed('write'))
        assert not permission_write.allows(identity_user_4)
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_user_3)
        assert current_access.get_action_cache('write') == (
            set([Need(method='id', value=3)]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2)]),
            set([])
        )

        # If we remove a user from an action, the cache should clear the
        # action item.
        user_3_action_write = ActionUsers.query.filter(
            ActionUsers.action == 'write' and
            ActionUsers.user == user_3).first()
        db.session.delete(user_3_action_write)
        db.session.flush()
        assert current_access.get_action_cache('write') is None
        # If no one is allowed to perform an action then everybody is allowed.
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_user_3)
        assert current_access.get_action_cache('write') == (
            set([]),
            set([])
        )
        db.session.add(ActionUsers(action='write', user=user_5))
        db.session.flush()
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_user_5)
        permission_write = dynamic_permission(ActionNeed('write'))
        assert not permission_write.allows(identity_user_3)
        assert current_access.get_action_cache('write') == (
            set([Need(method='id', value=5)]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2)]),
            set([])
        )

        # If you update the name of an existing action, the previous action
        # and the new action should be remove from cache.
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_user_5)
        assert current_access.get_action_cache('write') == (
            set([Need(method='id', value=5)]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2)]),
            set([])
        )
        user_5_action_write = ActionUsers.query.filter(
            ActionUsers.action == 'write' and
            ActionUsers.user == user_5).first()
        user_5_action_write.action = 'open'
        db.session.flush()
        assert current_access.get_action_cache('write') is None
        assert current_access.get_action_cache('open') is None
        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_user_1)
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2),
                 Need(method='id', value=5)]),
            set([])
        )
        db.session.add(ActionUsers(action='write', user=user_4))
        permission_write = dynamic_permission(ActionNeed('write'))
        assert not permission_write.allows(identity_user_5)
        assert current_access.get_action_cache('write') == (
            set([Need(method='id', value=4)]),
            set([])
        )

        db.session.add(ActionUsers(action='open', argument='1', user=user_6))
        db.session.flush()
        permission_open_1 = dynamic_permission(
            ParameterizedActionNeed('open', '1'))
        assert not permission_open.allows(identity_user_6)
        assert permission_open_1.allows(identity_user_6)
        assert current_access.get_action_cache('open::1') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2),
                 Need(method='id', value=5),
                 Need(method='id', value=6)]),
            set([])
        )
        user_6_action_open_1 = ActionUsers.query.filter_by(
            action='open', argument='1', user_id=user_6.id).first()
        user_6_action_open_1.argument = '2'
        db.session.flush()
        assert current_access.get_action_cache('open::1') is None
        assert current_access.get_action_cache('open::2') is None
        permission_open_2 = dynamic_permission(
            ParameterizedActionNeed('open', '2'))
        assert permission_open_2.allows(identity_user_6)
        assert current_access.get_action_cache('open::2') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2),
                 Need(method='id', value=5),
                 Need(method='id', value=6)]),
            set([])
        )
        # open action cache should remain as before
        assert current_access.get_action_cache('open') == (
            set([Need(method='id', value=1),
                 Need(method='id', value=2),
                 Need(method='id', value=5)]),
            set([])
        )


def test_invenio_access_permission_cache_roles_updates(
    app,
    dynamic_permission
):
    """Testing ActionRoles cache with inserts/updates/deletes."""
    # This test case is doing the same of user test case but using roles.
    cache = SimpleCache()
    InvenioAccess(app, cache=cache)
    with app.test_request_context():
        # Creation of some data to test.
        role_1 = Role(name='role_1')
        role_2 = Role(name='role_2')
        role_3 = Role(name='role_3')
        role_4 = Role(name='role_4')
        role_5 = Role(name='role_5')
        role_6 = Role(name='role_6')

        db.session.add(role_1)
        db.session.add(role_2)
        db.session.add(role_3)
        db.session.add(role_4)
        db.session.add(role_5)
        db.session.add(role_6)

        db.session.add(ActionRoles(action='open', role=role_1))
        db.session.add(ActionRoles(action='write', role=role_4))

        db.session.flush()

        # Creation of identities to test.
        identity_fake_role_1 = FakeIdentity(RoleNeed(role_1.name))
        identity_fake_role_2 = FakeIdentity(RoleNeed(role_2.name))
        identity_fake_role_3 = FakeIdentity(RoleNeed(role_3.name))
        identity_fake_role_4 = FakeIdentity(RoleNeed(role_4.name))
        identity_fake_role_5 = FakeIdentity(RoleNeed(role_5.name))
        identity_fake_role_6 = FakeIdentity(RoleNeed(role_6.name))

        # Test if role 1 can open. In this case, the cache should store only
        # this object.
        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_fake_role_1)
        assert current_access.get_action_cache('open') == (
            set([Need(method='role', value=role_1.name)]),
            set([])
        )

        # Test if role 4 can write. In this case, the cache should have this
        # new object and the previous one (Open is allowed to role_1)
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_role_4)
        assert current_access.get_action_cache('write') == (
            set([Need(method='role', value=role_4.name)]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([Need(method='role', value=role_1.name)]),
            set([])
        )

        # If we add a new role to the action open, the open action in cache
        # should be removed but it should still containing the write entry.
        db.session.add(ActionRoles(action='open', role=role_2))
        db.session.flush()
        assert current_access.get_action_cache('open') is None
        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_fake_role_2)
        assert current_access.get_action_cache('open') == (
            set([Need(method='role', value=role_1.name),
                 Need(method='role', value=role_2.name)]),
            set([])
        )
        assert current_access.get_action_cache('write') == (
            set([Need(method='role', value=role_4.name)]),
            set([])
        )

        # Test if the new role is added to the action 'open'
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_role_4)
        assert current_access.get_action_cache('open') == (
            set([Need(method='role', value=role_1.name),
                 Need(method='role', value=role_2.name)]),
            set([])
        )
        assert current_access.get_action_cache('write') == (
            set([Need(method='role', value=role_4.name)]),
            set([])
        )

        # If we update an action swapping a role, the cache containing the
        # action, should be removed.
        role_4_action_write = ActionRoles.query.filter(
            ActionRoles.action == 'write' and
            ActionRoles.role == role_4).first()
        role_4_action_write.role = role_3
        db.session.flush()

        assert current_access.get_action_cache('write') is None
        assert current_access.get_action_cache('open') is not None
        assert current_access.get_action_cache('open') == (
            set([Need(method='role', value=role_1.name),
                 Need(method='role', value=role_2.name)]),
            set([])
        )

        # Test if the role_3 can write now.
        permission_write = dynamic_permission(ActionNeed('write'))
        assert not permission_write.allows(identity_fake_role_4)
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_role_3)
        assert current_access.get_action_cache('write') == (
            set([Need(method='role', value=role_3.name)]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([Need(method='role', value=role_1.name),
                 Need(method='role', value=role_2.name)]),
            set([])
        )

        # If we remove a role from an action, the cache should clear the
        # action item.
        role_3_action_write = ActionRoles.query.filter(
            ActionRoles.action == 'write' and
            ActionRoles.role == role_3).first()
        db.session.delete(role_3_action_write)
        db.session.flush()
        assert current_access.get_action_cache('write') is None
        # If no one is allowed to perform an action then everybody is allowed.
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_role_3)
        assert current_access.get_action_cache('write') == (
            set([]),
            set([])
        )
        db.session.add(ActionRoles(action='write', role=role_5))
        db.session.flush()
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_role_5)
        permission_write = dynamic_permission(ActionNeed('write'))
        assert not permission_write.allows(identity_fake_role_3)
        assert current_access.get_action_cache('write') == (
            set([Need(method='role', value=role_5.name)]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([Need(method='role', value=role_1.name),
                 Need(method='role', value=role_2.name)]),
            set([])
        )

        # If you update the name of an existing action, the previous action
        # and the new action should be remove from cache.
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_role_5)
        assert current_access.get_action_cache('write') == (
            set([Need(method='role', value=role_5.name)]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([Need(method='role', value=role_1.name),
                 Need(method='role', value=role_2.name)]),
            set([])
        )
        role_5_action_write = ActionRoles.query.filter(
            ActionRoles.action == 'write' and
            ActionRoles.role == role_5).first()
        role_5_action_write.action = 'open'
        db.session.flush()
        assert current_access.get_action_cache('write') is None
        assert current_access.get_action_cache('open') is None
        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_fake_role_1)
        assert current_access.get_action_cache('open') == (
            set([Need(method='role', value=role_1.name),
                 Need(method='role', value=role_2.name),
                 Need(method='role', value=role_5.name)]),
            set([])
        )
        db.session.add(ActionRoles(action='write', role=role_4))
        permission_write = dynamic_permission(ActionNeed('write'))
        assert not permission_write.allows(identity_fake_role_5)
        assert current_access.get_action_cache('write') == (
            set([Need(method='role', value=role_4.name)]),
            set([])
        )

        db.session.add(ActionRoles(action='open', argument='1', role=role_6))
        db.session.flush()
        permission_open_1 = dynamic_permission(
            ParameterizedActionNeed('open', '1'))
        assert not permission_open.allows(identity_fake_role_6)
        assert permission_open_1.allows(identity_fake_role_6)
        assert current_access.get_action_cache('open::1') == (
            set([Need(method='role', value=role_1.name),
                 Need(method='role', value=role_2.name),
                 Need(method='role', value=role_5.name),
                 Need(method='role', value=role_6.name)]),
            set([])
        )
        user_6_action_open_1 = ActionRoles.query.filter_by(
            action='open', argument='1', role_id=role_6.id).first()
        user_6_action_open_1.argument = '2'
        db.session.flush()
        assert current_access.get_action_cache('open::1') is None
        assert current_access.get_action_cache('open::2') is None
        permission_open_2 = dynamic_permission(
            ParameterizedActionNeed('open', '2'))
        assert permission_open_2.allows(identity_fake_role_6)
        assert current_access.get_action_cache('open::2') == (
            set([Need(method='role', value=role_1.name),
                 Need(method='role', value=role_2.name),
                 Need(method='role', value=role_5.name),
                 Need(method='role', value=role_6.name)]),
            set([])
        )
        # open action cache should remain as before
        assert current_access.get_action_cache('open') == (
            set([Need(method='role', value=role_1.name),
                 Need(method='role', value=role_2.name),
                 Need(method='role', value=role_5.name)]),
            set([])
        )


def test_invenio_access_permission_cache_system_roles_updates(
    app,
    dynamic_permission
):
    """Testing ActionSystemRoles cache with inserts/updates/deletes."""
    # This test case is doing the same of user test case but using
    # system roles.
    cache = SimpleCache()
    InvenioAccess(app, cache=cache)
    with app.test_request_context():
        system_role_1 = SystemRoleNeed('system_role_1')
        system_role_2 = SystemRoleNeed('system_role_2')
        system_role_3 = SystemRoleNeed('system_role_3')
        system_role_4 = SystemRoleNeed('system_role_4')
        system_role_5 = SystemRoleNeed('system_role_5')
        system_role_6 = SystemRoleNeed('system_role_6')
        current_access.system_roles = {
            'system_role_1': system_role_1,
            'system_role_2': system_role_2,
            'system_role_3': system_role_3,
            'system_role_4': system_role_4,
            'system_role_5': system_role_5,
            'system_role_6': system_role_6,
        }

        # Creation of some data to test.
        db.session.add(ActionSystemRoles(action='open',
                                         role_name=system_role_1.value))
        db.session.add(ActionSystemRoles(action='write',
                                         role_name=system_role_4.value))

        db.session.flush()

        # Creation of identities to test.
        identity_fake_1 = FakeIdentity(system_role_1)
        identity_fake_2 = FakeIdentity(system_role_2)
        identity_fake_3 = FakeIdentity(system_role_3)
        identity_fake_4 = FakeIdentity(system_role_4)
        identity_fake_5 = FakeIdentity(system_role_5)
        identity_fake_6 = FakeIdentity(system_role_6)

        # Test if system_role_1 can open. In this case, the cache should store
        # only this object.
        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_fake_1)
        assert current_access.get_action_cache('open') == (
            set([system_role_1]),
            set([])
        )

        # Test if system_role_2 can write. In this case, the cache should
        # have this new object and the previous one (Open is allowed to
        # system_role_1)
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_4)
        assert current_access.get_action_cache('write') == (
            set([system_role_4]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([system_role_1]),
            set([])
        )

        # If we add a new system role to the action open, the open action in
        # cache should be removed but it should still containing the write
        # entry.
        db.session.add(ActionSystemRoles(action='open',
                                         role_name=system_role_2.value))
        db.session.flush()
        assert current_access.get_action_cache('open') is None
        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_fake_2)
        assert current_access.get_action_cache('open') == (
            set([system_role_1, system_role_2]),
            set([])
        )
        assert current_access.get_action_cache('write') == (
            set([system_role_4]),
            set([])
        )

        # Test if the new role is added to the action 'open'
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_4)
        assert current_access.get_action_cache('open') == (
            set([system_role_1, system_role_2]),
            set([])
        )
        assert current_access.get_action_cache('write') == (
            set([system_role_4]),
            set([])
        )

        # If we update an action swapping a role, the cache containing the
        # action, should be removed.
        role_4_action_write = ActionSystemRoles.query.filter(
            ActionSystemRoles.action == 'write' and
            ActionSystemRoles.role_name == system_role_4.value).first()
        role_4_action_write.role_name = system_role_3.value
        db.session.flush()

        assert current_access.get_action_cache('write') is None
        assert current_access.get_action_cache('open') is not None
        assert current_access.get_action_cache('open') == (
            set([system_role_1, system_role_2]),
            set([])
        )

        # Test if the system_role_3 can write now.
        permission_write = dynamic_permission(ActionNeed('write'))
        assert not permission_write.allows(identity_fake_4)
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_3)
        assert current_access.get_action_cache('write') == (
            set([system_role_3]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([system_role_1, system_role_2]),
            set([])
        )

        # If we remove a role from an action, the cache should clear the
        # action item.
        cust_action_write = ActionSystemRoles.query.filter(
            ActionSystemRoles.action == 'write' and
            ActionSystemRoles.role_name == system_role_3).first()
        db.session.delete(cust_action_write)
        db.session.flush()
        assert current_access.get_action_cache('write') is None
        # If no one is allowed to perform an action then everybody is allowed.
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_3)
        assert current_access.get_action_cache('write') == (
            set([]),
            set([])
        )
        db.session.add(ActionSystemRoles(action='write',
                                         role_name=system_role_5.value))
        db.session.flush()
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_5)
        permission_write = dynamic_permission(ActionNeed('write'))
        assert not permission_write.allows(identity_fake_3)
        assert current_access.get_action_cache('write') == (
            set([system_role_5]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([system_role_1, system_role_2]),
            set([])
        )

        # If you update the name of an existing action, the previous action
        # and the new action should be remove from cache.
        permission_write = dynamic_permission(ActionNeed('write'))
        assert permission_write.allows(identity_fake_5)
        assert current_access.get_action_cache('write') == (
            set([system_role_5]),
            set([])
        )
        assert current_access.get_action_cache('open') == (
            set([system_role_1, system_role_2]),
            set([])
        )
        role_5_action_write = ActionSystemRoles.query.filter(
            ActionSystemRoles.action == 'write' and
            ActionSystemRoles.role_name == system_role_5.value).first()
        role_5_action_write.action = 'open'
        db.session.flush()
        assert current_access.get_action_cache('write') is None
        assert current_access.get_action_cache('open') is None
        permission_open = dynamic_permission(ActionNeed('open'))
        assert permission_open.allows(identity_fake_1)
        assert current_access.get_action_cache('open') == (
            set([system_role_1, system_role_2, system_role_5]),
            set([])
        )
        db.session.add(ActionSystemRoles(action='write',
                                         role_name=system_role_4.value))
        permission_write = dynamic_permission(ActionNeed('write'))
        assert not permission_write.allows(identity_fake_5)
        assert current_access.get_action_cache('write') == (
            set([system_role_4]),
            set([])
        )

        db.session.add(ActionSystemRoles(action='open', argument='1',
                                         role_name=system_role_6.value))
        db.session.flush()
        permission_open_1 = dynamic_permission(
            ParameterizedActionNeed('open', '1'))
        assert not permission_open.allows(identity_fake_6)
        assert permission_open_1.allows(identity_fake_6)
        assert current_access.get_action_cache('open::1') == (
            set([system_role_1, system_role_2, system_role_5, system_role_6]),
            set([])
        )
        user_6_action_open_1 = ActionSystemRoles.query.filter_by(
            action='open', argument='1', role_name=system_role_6.value).first()
        user_6_action_open_1.argument = '2'
        db.session.flush()
        assert current_access.get_action_cache('open::1') is None
        assert current_access.get_action_cache('open::2') is None
        permission_open_2 = dynamic_permission(
            ParameterizedActionNeed('open', '2'))
        assert permission_open_2.allows(identity_fake_6)
        assert current_access.get_action_cache('open::2') == (
            set([system_role_1, system_role_2, system_role_5, system_role_6]),
            set([])
        )
        # open action cache should remain as before
        assert current_access.get_action_cache('open') == (
            set([system_role_1, system_role_2, system_role_5]),
            set([])
        )


def test_dynamic_permission_needs_cache_invalidation(app, dynamic_permission):
    """Testing DynamicPermission refreshes needs.

    This is important when protecting a view with
    @permission.require(http_exception=403)
    If cache does not get invalidated, the needs will only be refreshed when
    the Python process restarts.
    """
    cache = SimpleCache()
    InvenioAccess(app, cache=cache)
    with app.test_request_context():
        user_can_all = User(email='all@inveniosoftware.org')
        user_can_open = User(email='open@inveniosoftware.org')
        db.session.add(user_can_all)
        db.session.add(user_can_open)

        db.session.add(ActionUsers(action='open', user=user_can_all))
        db.session.flush()

        permission_open = dynamic_permission(ActionNeed('open'))

        assert permission_open.needs == set([Need(method='id', value=1)])

        db.session.add(ActionUsers(action='open', user=user_can_open))
        db.session.flush()

        assert permission_open.needs == set(
            [Need(method='id', value=1), Need(method='id', value=2)]
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Database models for access module."""

from __future__ import absolute_import, print_function

from flask_principal import RoleNeed, UserNeed
from invenio_accounts.models import Role, User
from invenio_db import db
from sqlalchemy import UniqueConstraint
from sqlalchemy.event import listen
from sqlalchemy.orm import validates
from sqlalchemy.orm.attributes import get_history

from .proxies import current_access


class ActionNeedMixin(object):
    """Define common attributes for Action needs."""

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    """Primary key. It allows the other fields to be nullable."""

    action = db.Column(db.String(80), index=True)
    """Name of the action."""

    exclude = db.Column(db.Boolean(name='exclude'), nullable=False,
                        default=False, server_default='0')
    """If set to True, deny the action, otherwise allow it."""

    argument = db.Column(db.String(255), nullable=True, index=True)
    """Action argument."""

    @classmethod
    def create(cls, action, **kwargs):
        """Create new database row using the provided action need.

        :param action: An object containing a method equal to ``'action'`` and
            a value.
        :param argument: The action argument. If this parameter is not passed,
            then the ``action.argument`` will be used instead. If the
            ``action.argument`` does not exist, ``None`` will be set as
            argument for the new action need.
        :returns: An :class:`invenio_access.models.ActionNeedMixin` instance.
        """
        assert action.method == 'action'
        argument = kwargs.pop('argument', None) or getattr(
            action, 'argument', None)
        return cls(
            action=action.value,
            argument=argument,
            **kwargs
        )

    @classmethod
    def allow(cls, action, **kwargs):
        """Allow the given action need.

        :param action: The action to allow.
        :returns: A :class:`invenio_access.models.ActionNeedMixin` instance.
        """
        return cls.create(action, exclude=False, **kwargs)

    @classmethod
    def deny(cls, action, **kwargs):
        """Deny the given action need.

        :param action: The action to deny.
        :returns: A :class:`invenio_access.models.ActionNeedMixin` instance.
        """
        return cls.create(action, exclude=True, **kwargs)

    @classmethod
    def query_by_action(cls, action, argument=None):
        """Prepare query object with filtered action.

        :param action: The action to deny.
        :param argument: The action argument. If it's ``None`` then, if exists,
            the ``action.argument`` will be taken. In the worst case will be
            set as ``None``. (Default: ``None``)
        :returns: A query object.
        """
        query = cls.query.filter_by(action=action.value)
        argument = argument or getattr(action, 'argument', None)
        if argument is not None:
            query = query.filter(db.or_(
                cls.argument == str(argument),
                cls.argument.is_(None),
            ))
        else:
            query = query.filter(cls.argument.is_(None))
        return query

    @property
    def need(self):
        """Return the need corresponding to this model instance.

        This is an abstract method and will raise NotImplementedError.
        """
        raise NotImplementedError()  # pragma: no cover


class ActionUsers(ActionNeedMixin, db.Model):
    """ActionUsers data model.

    It relates an allowed action with a user.
    """

    __tablename__ = 'access_actionsusers'

    __table_args__ = (UniqueConstraint(
        'action', 'exclude', 'argument', 'user_id',
        name='access_actionsusers_unique'),
    )

    user_id = db.Column(db.Integer(),
                        db.ForeignKey(User.id, ondelete='CASCADE'),
                        nullable=False, index=True)

    user = db.relationship("User",
                           backref=db.backref("actionusers",
                                              cascade="all, delete-orphan"))

    @property
    def need(self):
        """Return UserNeed instance."""
        return UserNeed(self.user_id)


class ActionRoles(ActionNeedMixin, db.Model):
    """ActionRoles data model.

    It relates an allowed action with a role.
    """

    __tablename__ = 'access_actionsroles'

    __table_args__ = (UniqueConstraint(
        'action', 'exclude', 'argument', 'role_id',
        name='access_actionsroles_unique'),
    )

    role_id = db.Column(db.Integer(),
                        db.ForeignKey(Role.id, ondelete='CASCADE'),
                        nullable=False, index=True)

    role = db.relationship("Role",
                           backref=db.backref("actionusers",
                                              cascade="all, delete-orphan"))

    @property
    def need(self):
        """Return RoleNeed instance."""
        return RoleNeed(self.role.name)


class ActionSystemRoles(ActionNeedMixin, db.Model):
    """ActionSystemRoles data model.

    It relates an allowed action with a predefined role.
    Example: "any user"
    """

    __tablename__ = 'access_actionssystemroles'

    __table_args__ = (UniqueConstraint(
        'action', 'exclude', 'argument', 'role_name',
        name='access_actionssystemroles_unique'),
    )

    role_name = db.Column(db.String(40), nullable=False, index=True)

    @classmethod
    def create(cls, action, **kwargs):
        """Create new database row using the provided action need."""
        role = kwargs.pop('role', None)
        if role:
            assert role.method == 'system_role'
            kwargs['role_name'] = role.value
        return super(ActionSystemRoles, cls).create(action, **kwargs)

    @validates('role_name')
    def validate_role_name(self, key, role_name):
        """Checks that the role name has been registered."""
        assert role_name in current_access.system_roles
        return role_name

    @property
    def need(self):
        """Return the corresponding Need instance."""
        return current_access.system_roles[self.role_name]


def get_action_cache_key(name, argument):
    """Get an action cache key string."""
    tokens = [str(name)]
    if argument:
        tokens.append(str(argument))
    return '::'.join(tokens)


def removed_or_inserted_action(mapper, connection, target):
    """Remove the action from cache when an item is inserted or deleted."""
    current_access.delete_action_cache(get_action_cache_key(target.action,
                                                            target.argument))


def changed_action(mapper, connection, target):
    """Remove the action from cache when an item is updated."""
    action_history = get_history(target, 'action')
    argument_history = get_history(target, 'argument')
    owner_history = get_history(
        target,
        'user' if isinstance(target, ActionUsers) else
        'role' if isinstance(target, ActionRoles) else 'role_name')

    if action_history.has_changes() or argument_history.has_changes() \
       or owner_history.has_changes():
        current_access.delete_action_cache(
            get_action_cache_key(target.action, target.argument))
        current_access.delete_action_cache(
            get_action_cache_key(
                action_history.deleted[0] if action_history.deleted
                else target.action,
                argument_history.deleted[0] if argument_history.deleted
                else target.argument)
        )


listen(ActionUsers, 'after_insert', removed_or_inserted_action)
listen(ActionUsers, 'after_delete', removed_or_inserted_action)
listen(ActionUsers, 'after_update', changed_action)

listen(ActionRoles, 'after_insert', removed_or_inserted_action)
listen(ActionRoles, 'after_delete', removed_or_inserted_action)
listen(ActionRoles, 'after_update', changed_action)

listen(ActionSystemRoles, 'after_insert', removed_or_inserted_action)
listen(ActionSystemRoles, 'after_delete', removed_or_inserted_action)
listen(ActionSystemRoles, 'after_update', changed_action)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add on delete cascade constraint."""

from alembic import op

# revision identifiers, used by Alembic.
revision = '2069a982633b'
down_revision = '2f63be7b7572'
branch_labels = ()
depends_on = '35c1075e6360'  # invenio-db "Force naming convention"


def upgrade():
    """Upgrade database."""
    op.create_index(op.f('ix_access_actionsroles_role_id'),
                    'access_actionsroles', ['role_id'], unique=False)
    op.drop_constraint(u'fk_access_actionsroles_role_id_accounts_role',
                       'access_actionsroles', type_='foreignkey')
    op.create_foreign_key(op.f('fk_access_actionsroles_role_id_accounts_role'),
                          'access_actionsroles', 'accounts_role', ['role_id'],
                          ['id'], ondelete='CASCADE')
    op.create_index(op.f('ix_access_actionsusers_user_id'),
                    'access_actionsusers', ['user_id'], unique=False)
    op.drop_constraint(u'fk_access_actionsusers_user_id_accounts_user',
                       'access_actionsusers', type_='foreignkey')
    op.create_foreign_key(op.f('fk_access_actionsusers_user_id_accounts_user'),
                          'access_actionsusers', 'accounts_user', ['user_id'],
                          ['id'], ondelete='CASCADE')


def downgrade():
    """Downgrade database."""
    op.drop_constraint(op.f('fk_access_actionsusers_user_id_accounts_user'),
                       'access_actionsusers', type_='foreignkey')
    op.drop_index(op.f('ix_access_actionsusers_user_id'),
                  table_name='access_actionsusers')
    op.create_foreign_key(u'fk_access_actionsusers_user_id_accounts_user',
                          'access_actionsusers', 'accounts_user', ['user_id'],
                          ['id'])
    op.drop_constraint(op.f('fk_access_actionsroles_role_id_accounts_role'),
                       'access_actionsroles', type_='foreignkey')
    op.drop_index(op.f('ix_access_actionsroles_role_id'),
                  table_name='access_actionsroles')
    op.create_foreign_key(u'fk_access_actionsroles_role_id_accounts_role',
                          'access_actionsroles', 'accounts_role', ['role_id'],
                          ['id'])



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Command line interface for Invenio-Access."""

from __future__ import absolute_import, print_function

from functools import wraps

import click
from flask import current_app
from flask.cli import with_appcontext
from invenio_accounts.models import Role, User
from invenio_db import db
from werkzeug.local import LocalProxy

from .models import ActionRoles, ActionUsers

_current_actions = LocalProxy(
    lambda: current_app.extensions['invenio-access'].actions
)
"""Helper proxy to registered actions."""


def lazy_result(f):
    """Decorate function to return LazyProxy."""
    @wraps(f)
    def decorated(ctx, param, value):
        return LocalProxy(lambda: f(ctx, param, value))
    return decorated


@lazy_result
def process_action(ctx, param, value):
    """Return an action if exists."""
    actions = current_app.extensions['invenio-access'].actions
    if value not in actions:
        raise click.BadParameter('Action "%s" is not registered.', value)
    return actions[value]


@lazy_result
def process_email(ctx, param, value):
    """Return an user if it exists."""
    user = User.query.filter(User.email == value).first()
    if not user:
        raise click.BadParameter('User with email \'%s\' not found.', value)
    return user


@lazy_result
def process_role(ctx, param, value):
    """Return a role if it exists."""
    role = Role.query.filter(Role.name == value).first()
    if not role:
        raise click.BadParameter('Role with name \'%s\' not found.', value)
    return role


option_argument = click.option(
    '-a', '--argument', default=None, metavar='VALUE',
    help='Value for parameterized action.'
)
option_email = click.option(
    '-e', '--email', multiple=True, default=[], metavar='EMAIL',
    help='User email address(es).'
)
option_role = click.option(
    '-r', '--role', multiple=True, default=[], metavar='ROLE',
    help='Role name(s).'
)
argument_action = click.argument(
    'action', callback=process_action, nargs=1, required=True, metavar='ACTION'
)
argument_user = click.argument(
    'user', callback=process_email, nargs=1, required=True, metavar='EMAIL'
)
argument_role = click.argument(
    'role', callback=process_role, nargs=1, required=True, metavar='ROLE'
)


#
# Access commands
#
@click.group()
def access():
    """Account commands."""


#
# Allow Action
#
@access.group(name='allow', chain=True)
@argument_action
@option_argument
def allow_action(action, argument):
    """Allow action."""


@allow_action.command('user')
@argument_user
def allow_user(user):
    """Allow a user identified by an email address."""
    def processor(action, argument):
        db.session.add(
            ActionUsers.allow(action, argument=argument, user_id=user.id)
        )
    return processor


@allow_action.command('role')
@argument_role
def allow_role(role):
    """Allow a role identified by an email address."""
    def processor(action, argument):
        db.session.add(
            ActionRoles.allow(action, argument=argument, role_id=role.id)
        )
    return processor


@allow_action.resultcallback()
@with_appcontext
def process_allow_action(processors, action, argument):
    """Process allow action."""
    for processor in processors:
        processor(action, argument)
    db.session.commit()


#
# Deny Action
#
@access.group(name='deny', chain=True)
@argument_action
@option_argument
def deny_action(action, argument):
    """Deny actions."""


@deny_action.command('user')
@argument_user
def deny_user(user):
    """Deny a user identified by an email address."""
    def processor(action, argument):
        db.session.add(
            ActionUsers.deny(action, argument=argument, user_id=user.id)
        )
    return processor


@deny_action.command('role')
@argument_role
def deny_role(role):
    """Deny a role identified by an email address."""
    def processor(action, argument):
        db.session.add(
            ActionRoles.deny(action, argument=argument, role_id=role.id)
        )
    return processor


@deny_action.resultcallback()
@with_appcontext
def process_deny_action(processors, action, argument):
    """Process deny action."""
    for processor in processors:
        processor(action, argument)
    db.session.commit()


#
# Remove Action
#
@access.group(name='remove', chain=True)
@argument_action
@option_argument
def remove_action(action, argument):
    """Remove existing action authorization.

    It is possible to specify multiple emails and/or roles that
    should be unassigned from the given action.
    """


@remove_action.command('global')
def remove_global():
    """Remove global action rule."""
    def processor(action, argument):
        ActionUsers.query_by_action(action, argument=argument).filter(
            ActionUsers.user_id.is_(None)
        ).delete(synchronize_session=False)
    return processor


@remove_action.command('user')
@argument_user
def remove_user(user):
    """Remove a action for a user."""
    def processor(action, argument):
        ActionUsers.query_by_action(action, argument=argument).filter(
            ActionUsers.user_id == user.id
        ).delete(synchronize_session=False)
    return processor


@remove_action.command('role')
@argument_role
def remove_role(role):
    """Remove a action for a role."""
    def processor(action, argument):
        ActionRoles.query_by_action(action, argument=argument).filter(
            ActionRoles.role_id == role.id
        ).delete(synchronize_session=False)
    return processor


@remove_action.resultcallback()
@with_appcontext
def process_remove_action(processors, action, argument):
    """Process action removals."""
    for processor in processors:
        processor(action, argument)
    db.session.commit()


@access.command(name='list')
@with_appcontext
def list_actions():
    """List all registered actions."""
    for name, action in _current_actions.items():
        click.echo('{0}:{1}'.format(
            name, '*' if hasattr(action, 'argument') else ''
        ))


@access.command(name='show')
@option_email
@option_role
@with_appcontext
def show_actions(email, role):
    """Show all assigned actions."""
    if email:
        actions = ActionUsers.query.join(ActionUsers.user).filter(
            User.email.in_(email)
        ).all()
        for action in actions:
            click.secho('user:{0}:{1}:{2}:{3}'.format(
                action.user.email,
                action.action,
                '' if action.argument is None else action.argument,
                'deny' if action.exclude else 'allow',
            ), fg='red' if action.exclude else 'green')

    if role:
        actions = ActionRoles.query.filter(
            Role.name.in_(role)
        ).join(ActionRoles.role).all()
        for action in actions:
            click.secho('role:{0}:{1}:{2}:{3}'.format(
                action.role.name,
                action.action,
                '' if action.argument is None else action.argument,
                'deny' if action.exclude else 'allow',
            ), fg='red' if action.exclude else 'green')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tests for Permission class."""

import pytest
from flask_principal import ActionNeed, Need, RoleNeed, UserNeed
from invenio_accounts.models import Role, User
from invenio_db import db

from invenio_access.models import ActionRoles, ActionSystemRoles, ActionUsers
from invenio_access.permissions import ParameterizedActionNeed, Permission, \
    any_user, authenticated_user, superuser_access

# Test cases summary

# List of Need that will be tested:

# UserNeed
# RoleNeed
# ActionNeed
# ParameterizedActionNeed
# SystemRoleNeed

# Test cases (`->` means expand action to a user or a role):

# 1. allow UserNeed
# 2. deny UserNeed
# 3. allow RoleNeed
# 4. deny RoleNeed
# 5. allow UserNeed, RoleNeed
# 6. deny UserNeed, RoleNeed
# 7. allow ActionNeed -> UserNeed
# 8. allow ActionNeed -> RoleNeed
# 9. deny ActionNeed -> UserNeed
# 10. deny ActionNeed -> RoleNeed
# 11. allow ActionNeed -> UserNeed, RoleNeed
# 12. deny ActionNeed -> UserNeed, RoleNeed
# 13. allow/deny ParameterizedActionNeed
# 14. test SystemRoleNeed
# 15. allow_by_default = True/False


class WithUserIdentity(object):
    """Helper class to mock Flask identity and provide needs."""

    def __init__(self, user, *provides):
        self.user = user
        self.id = user.id
        self.provides = set(provides)
        self.provides.add(UserNeed(user.id))
        assert ActionNeed not in self.provides


def get_superuser():
    """Create a superuser user."""
    superuser, = create_users("superpowers")
    expand(superuser_access, ("allow", superuser))
    return superuser


def create_users(*names):
    """Helper to create users."""
    users = []
    for name in names:
        user = User(email="{}@inveniosoftware.org".format(name))
        db.session.add(user)
        users.append(user)
    db.session.commit()
    return [WithUserIdentity(user) for user in users]


def create_roles(*names):
    """Helper to create roles."""
    roles = []
    for name in names:
        role = Role(name=name)
        db.session.add(role)
        roles.append(role)
    db.session.commit()
    return roles


def assign_roles(users_roles):
    """Assign roles to users."""
    for user, roles in users_roles.items():
        for role in roles:
            if not isinstance(role, Need):
                role = RoleNeed(role.name)
            user.provides.add(role)


def create_permissions(*permissions):
    """Create permissions with given needs and excludes."""
    _permissions = []
    for _needs in permissions:
        permission = Permission()
        permission.explicit_needs.update(_needs.get("needs", set()))
        permission.explicit_excludes.update(_needs.get("excludes", set()))
        _permissions.append(permission)
    return _permissions


def expand(action_need, *user_or_roles):
    """Expand action to allow/deny users or roles."""
    for user_or_role in user_or_roles:
        allow_deny = user_or_role[0]
        user_role = user_or_role[1]
        argument = user_or_role[2] if len(user_or_role) == 3 else None
        is_user = isinstance(user_role, WithUserIdentity)
        if is_user:
            # User
            kwargs = dict(user=user_role.user)
            method = (
                ActionUsers.allow
                if allow_deny == "allow"
                else ActionUsers.deny
            )
        elif isinstance(user_role, Need) and user_role.method == "system_role":
            # SystemRoleNeed
            kwargs = dict(role=user_role)
            method = (
                ActionSystemRoles.allow
                if allow_deny == "allow"
                else ActionSystemRoles.deny
            )
        else:
            # Role
            kwargs = dict(role=user_role)
            method = (
                ActionRoles.allow
                if allow_deny == "allow"
                else ActionRoles.deny
            )

        obj = method(action=action_need, argument=argument, **kwargs)
        db.session.add(obj)
    db.session.commit()
    return action_need


def test_allow_by_user_id(access_app):
    """Ensure allow permission by user id."""
    admin, reader = create_users("admin", "reader")

    permission, = create_permissions({"needs": {UserNeed(admin.id)}})

    assert permission.allows(admin)
    assert permission.allows(get_superuser())
    assert not permission.allows(reader)


def test_deny_by_user_id(access_app):
    """Ensure deny permission when user is denied."""
    admin, reader = create_users("admin", "reader")

    permission, = create_permissions(
        {"excludes": [UserNeed(admin.id), UserNeed(reader.id)]}
    )

    assert not permission.allows(admin)
    assert not permission.allows(reader)
    assert permission.allows(get_superuser())


def test_allow_deny_by_user_id(access_app):
    """Ensure allow/deny permission when same user is allowed and denied."""
    admin, reader = create_users("admin", "reader")

    permission, = create_permissions(
        {
            "needs": [UserNeed(admin.id)],
            "excludes": [UserNeed(admin.id), UserNeed(reader.id)],
        }
    )

    # `excludes` prevail over `needs`
    assert not permission.allows(admin)
    assert not permission.allows(reader)
    assert permission.allows(get_superuser())


def test_allow_by_role(access_app):
    """Ensure allow permission by role."""
    admin, reader = create_users("admin", "reader")
    administrators, readers = create_roles("administrators", "readers")
    assign_roles({admin: [administrators], reader: [readers]})

    permission, = create_permissions(
        {"needs": [RoleNeed(administrators.name)]}
    )

    assert permission.allows(admin)
    assert not permission.allows(reader)
    assert permission.allows(get_superuser())


def test_deny_by_role(access_app):
    """Ensure deny permission by role."""
    admin, reader = create_users("admin", "reader")
    administrators, readers = create_roles("administrators", "readers")
    assign_roles({admin: [administrators], reader: [readers]})

    permission, = create_permissions(
        {
            "needs": [RoleNeed(administrators.name)],
            "excludes": [
                RoleNeed(administrators.name),
                RoleNeed(readers.name),
            ],
        }
    )

    # `excludes` prevail over `needs`
    assert not permission.allows(admin)
    assert not permission.allows(reader)
    assert permission.allows(get_superuser())


def test_allow_by_user_id_and_role(access_app):
    """Ensure allow permission by user id and role."""
    admin, reader = create_users("admin", "reader")
    administrators, readers = create_roles("administrators", "readers")
    assign_roles({admin: [administrators], reader: [readers]})

    permission, = create_permissions(
        {"needs": [UserNeed(reader.id), RoleNeed(administrators.name)]}
    )

    assert permission.allows(admin)
    assert permission.allows(reader)
    assert permission.allows(get_superuser())


def test_deny_by_user_id_and_role(access_app):
    """Ensure deny permission by user id and role."""
    admin, reader = create_users("admin", "reader")
    administrators, readers, any_user = create_roles(
        "administrators", "readers", "any_user"
    )
    assign_roles(
        {admin: [administrators, any_user], reader: [readers, any_user]}
    )

    permission, = create_permissions(
        {
            "needs": [UserNeed(reader.id), RoleNeed(administrators.name)],
            "excludes": [RoleNeed(any_user.name)],
        }
    )

    assert not permission.allows(admin)
    assert not permission.allows(reader)
    assert permission.allows(get_superuser())


def test_allow_by_action_expands_to_user_id(access_app):
    """Ensure allow permission by action that expands to user id."""
    admin, reader = create_users("admin", "reader")
    administrators, readers = create_roles("administrators", "readers")
    assign_roles({admin: [administrators], reader: [readers]})

    act_access_backoffice = expand(
        ActionNeed("access-backoffice"), ("allow", admin)
    )
    permission, = create_permissions({"needs": [act_access_backoffice]})

    assert permission.allows(admin)
    assert not permission.allows(reader)
    assert permission.allows(get_superuser())


def test_deny_by_action_expands_to_user_id(access_app):
    """Ensure deny permission by action that expands to user id."""
    admin, reader = create_users("admin", "reader")
    administrators, readers = create_roles("administrators", "readers")
    assign_roles({admin: [administrators], reader: [readers]})

    act_access_backoffice = expand(
        ActionNeed("access-backoffice"),
        ("allow", admin),
        ("allow", reader),
        ("deny", admin),
        ("deny", reader),
    )
    permission, = create_permissions({"needs": [act_access_backoffice]})

    assert not permission.allows(admin)
    assert not permission.allows(reader)
    assert permission.allows(get_superuser())


def test_allow_by_action_expands_to_role(access_app):
    """Ensure allow permission by action that expands to role."""
    admin, reader = create_users("admin", "reader")
    administrators, readers = create_roles("administrators", "readers")
    assign_roles({admin: [administrators], reader: [readers]})

    act_access_backoffice = expand(
        ActionNeed("access-backoffice"), ("allow", administrators)
    )
    permission, = create_permissions({"needs": [act_access_backoffice]})

    assert permission.allows(admin)
    assert not permission.allows(reader)
    assert permission.allows(get_superuser())


def test_deny_by_action_expands_to_role(access_app):
    """Ensure deny permission by action that expands to role."""
    admin, reader = create_users("admin", "reader")
    administrators, readers = create_roles("administrators", "readers")
    assign_roles({admin: [administrators], reader: [readers]})

    act_access_backoffice = expand(
        ActionNeed("access-backoffice"),
        ("allow", administrators),
        ("allow", readers),
        ("deny", administrators),
        ("deny", readers),
    )
    permission, = create_permissions({"needs": [act_access_backoffice]})

    assert not permission.allows(admin)
    assert not permission.allows(reader)
    assert permission.allows(get_superuser())


def test_allow_by_action_expands_to_user_id_and_role(access_app):
    """Ensure allow permission by action that expands to user id and role."""
    admin, reader = create_users("admin", "reader")
    administrators, readers = create_roles("administrators", "readers")
    assign_roles({admin: [administrators], reader: [readers]})

    act_access_backoffice = expand(
        ActionNeed("access-backoffice"),
        ("allow", administrators),
        ("allow", reader),
    )
    permission, = create_permissions({"needs": [act_access_backoffice]})

    assert permission.allows(admin)
    assert permission.allows(reader)
    assert permission.allows(get_superuser())


def test_deny_by_action_expands_to_user_id_and_role(access_app):
    """Ensure deny permission by action that expands to user id and role."""
    admin, reader = create_users("admin", "reader")
    administrators, readers = create_roles("administrators", "readers")
    assign_roles({admin: [administrators], reader: [readers]})

    act_access_backoffice = expand(
        ActionNeed("access-backoffice"),
        ("allow", administrators),
        ("allow", reader),
        ("deny", administrators),
        ("deny", reader),
    )
    permission, = create_permissions({"needs": [act_access_backoffice]})

    assert not permission.allows(admin)
    assert not permission.allows(reader)
    assert permission.allows(get_superuser())


def test_allow_by_action_expands_to_user_id_with_argument(access_app):
    """Ensure allow permission by action that expands to user id + argument."""
    editor1, editor2 = create_users("editor1", "editor2")

    act_edit_record_1 = expand(
        ParameterizedActionNeed("access-edit-record", "1"),
        ("allow", editor1, "1"),
    )
    act_edit_record_2 = expand(
        ParameterizedActionNeed("access-edit-record", "2"),
        ("allow", editor2, "2"),
    )
    permission1, permission2 = create_permissions(
        {"needs": [act_edit_record_1]}, {"needs": [act_edit_record_2]}
    )

    assert permission1.allows(editor1)
    assert not permission1.allows(editor2)
    assert permission2.allows(editor2)
    assert not permission2.allows(editor1)

    superuser = get_superuser()
    assert permission1.allows(superuser)
    assert permission2.allows(superuser)


def test_system_role_name(access_app):
    """Test that ActionSystemRoles fails when created with undeclared names."""
    state = access_app.extensions["invenio-access"]
    db.session.begin(nested=True)
    # Declare a system role.
    state.system_roles = {"any_user": any_user}
    # Create with a declared name.
    ActionSystemRoles(action="read", role_name="any_user")
    # Create with an undeclared name.
    with pytest.raises(AssertionError):
        ActionSystemRoles(action="read", role_name="unknown")


def test_system_roles(access_app):
    """User can access to an action allowed/denied to their system roles."""
    authenticated, anonymous = create_users("authenticated", "anonymous")
    assign_roles(
        {authenticated: [any_user, authenticated_user], anonymous: [any_user]}
    )

    act_read = ActionNeed("read")
    act_write = ActionNeed("write")

    expand(act_write, ("allow", authenticated_user))
    expand(act_read, ("allow", any_user))

    permission_read, permission_write = create_permissions(
        {"needs": [act_read]}, {"needs": [act_write]}
    )

    assert permission_read.allows(anonymous)
    assert permission_read.allows(authenticated)
    assert not permission_write.allows(anonymous)
    assert permission_write.allows(authenticated)

    superuser = get_superuser()
    assert permission_read.allows(superuser)
    assert permission_write.allows(superuser)


def test_allow_by_default(access_app, dynamic_permission):
    """Test that dynamic permissions allows access by default."""
    anonymous, superuser = create_users("anonymous", "superuser")
    act_access_backoffice = ActionNeed("access-backoffice")

    permission, = create_permissions({"needs": [act_access_backoffice]})
    dyn_permission = dynamic_permission(act_access_backoffice)

    assert not permission.allows(anonymous)
    assert dyn_permission.allows(anonymous)

    superuser = get_superuser()
    assert permission.allows(superuser)
    assert dyn_permission.allows(superuser)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Role-based access control for Invenio.

Invenio-Access works together with Invenio-Accounts to provide a full-fledge
authentication and authorization system for Flask and Invenio based on a suite
of existing Flask extensions such as:

- Flask-Security
- Flask-Login
- Flask-Principal
- passlib

Make sure you check out :ref:`concepts` to have a basic understanding of the
entities in the access control system. This part of the usage documentation is
focused on the programmatic APIs and are intended for developers.

Initialization
--------------

Create a Flask application:

>>> import os
>>> db_url = os.environ.get('SQLALCHEMY_DATABASE_URI', 'sqlite://')
>>> from flask import Flask
>>> app = Flask('myapp')
>>> app.config.update({
...     'SQLALCHEMY_DATABASE_URI': db_url,
...     'SQLALCHEMY_TRACK_MODIFICATIONS': False,
... })

Initialize Invenio-Access dependencies, which are Invenio-DB and
Invenio-Accounts, and then Invenio-Access itself:

>>> from invenio_db import InvenioDB
>>> from invenio_accounts import InvenioAccounts
>>> from invenio_access import InvenioAccess
>>> ext_db = InvenioDB(app)
>>> ext_accounts = InvenioAccounts(app)
>>> ext_access = InvenioAccess(app)

The following examples needs to run in a Flask application context, so
let's push one:

>>> app.app_context().push()

Also, for the examples to work we need to create the database and tables (note,
in this example we use an in-memory SQLite database by default):

>>> from invenio_db import db
>>> db.create_all()

Demo data
~~~~~~~~~
Let's also create two initial users and a role:

>>> from invenio_accounts.models import User, Role
>>> alice = User(email='alice@inveniosoftware.org')
>>> bob = User(email='bob@inveniosoftware.org')
>>> admin = Role(name='admin')

Now, assign Alice to the admin role:

>>> admin.users.append(alice)

Last, persist the changees to the database:

>>> db.session.add_all([alice, bob, admin])
>>> db.session.commit()

Protecting resources
--------------------
The basics of protecting a resource involves:

1. Defining one or more actions.
2. Create a permission that requires one or more actions.
3. Check if a permission allows a given identity (i.e. the identity provides
   one more of the required actions).

**1. Define an action**

First, let's start with defining an action (e.g. view an index page in our
module) using the action creation factory:

>>> from invenio_access import action_factory
>>> view_index_action = action_factory('mymodule-index-view')

**2. Create a permission**

Next, we create a permission that requires the just created action:

>>> from invenio_access import Permission
>>> permission = Permission(view_index_action)

**3. Check permission**

In order to check the permission we first need an identity, so let's start out
with an anonymous identity (this happens transparently in the background when
a user login):

>>> from flask_principal import AnonymousIdentity
>>> anonymous = AnonymousIdentity()

Next, we can check if the permission allows the given identity (we will see
in detail below how to use permissions in a view):

>>> permission.allows(anonymous)
False

Granting access
---------------
Checking if the anonymous identity is granted access by a permission is often
not too interesting, so let's grant our admin role access to our action:

>>> from invenio_access.models import ActionRoles, ActionUsers
>>> db.session.add(ActionRoles.allow(view_index_action, role=admin))
>>> db.session.commit()

Next, we need identity instances for our two users (normally you will not have
to worry about this when checking permissions in a view as it is handled
transparently by Flask-Security):

>>> from invenio_access.utils import get_identity
>>> alice_identity = get_identity(alice)
>>> bob_identity = get_identity(bob)

Now that we have the identities, we can check if the permission grants access
to the identities:

>>> permission.allows(alice_identity)
True
>>> permission.allows(bob_identity)
False

Notice, that we granted access to Alice by assigning her the role ``admin``
and granting the role permission on the action.

The Flask-Principal API is pretty rich, and there are multiple other ways
that you can check if a permission grants access to an identity. For instance
below is another example, but please explore the Flask-Principal API
documentation for a full reference:

>>> bob_identity.can(permission)
False

Action parameters
-----------------
Above we created an action that did not take any parameters. These actions are
useful to grant/restrict access to e.g. an entire adminstration interface.
However, in many cases you need object level permissions, in which case you
need to use actions with parameters.

Action with parameters are also created with the
:py:func:`~.factory.action_factory`, but works a bit different as
they take a parameter.

First you create the new action with parameter:

>>> ObjectReadAction = action_factory(
...     'mymodule-object-read', parameter=True)

Everytime you create the action, you also need to create an instance of the
action representing *any* parameter as done like below:

>>> object_read_action_all = ObjectReadAction(None)

**Granting access to actions with parameters**

You grant access to actions with parameters in a similar way as for normal
actions, but you can now grant access to either **all** objects:

>>> db.session.add(ActionRoles.allow(
...     object_read_action_all, role=admin))

Or you can grant access to a **specific** object like this:

>>> db.session.add(ActionUsers.allow(ObjectReadAction(42), user=bob))
>>> db.session.commit()

**Checking permissions for a specific object**

Similar you also create a permission that checks access to a specific object:

>>> permission = Permission(ObjectReadAction(42))
>>> permission.allows(bob_identity)
True
>>> permission.allows(alice_identity)
True

Denying access
--------------
Besides granting access, you can also **deny** access to specific users or
roles. Below for instance we deny access to Alice on the ``view_index_action``.

>>> from invenio_access.models import ActionUsers
>>> db.session.add(ActionUsers.deny(view_index_action, user=alice))
>>> db.session.commit()

When we now check the permission, Alice no longer has access:

>>> permission = Permission(view_index_action)
>>> permission.allows(alice_identity)
False

**Deny takes precedence over allow**

Note, that the deny grant takes precedence over allow grant. Alice was earlier
granted access to the action via her role assignment, however since the deny
grant takes precedence Alice is ultimately denied access.

This is useful if you for instance want to grant access to all objects
*except* one.

Protecting views
----------------
The most common use for permissions is to protect a view. For actions without
parameters you can simply use a decorator for the view:

>>> index_permission = Permission(view_index_action)
>>> @app.route('/')
... @index_permission.require(http_exception=403)
... def index():
...     return 'Protected index page'

Permission factories
~~~~~~~~~~~~~~~~~~~~
In most situations, you however have to deal with object level permissions,
and thus you will have to create the permission on-the-fly via a factory
method. A simple permission factory can look like the one below:

>>> def permission_factory(obj):
...     return Permission(ObjectReadAction(obj['id']))

The factory function simply takes your object and returns a permission for the
specific action. This unfortunately also means that you usually cannot use
the decorator option shown above. Instead you usually have to first fetch your
object from e.g. the database, and then run the permission check:

>>> @app.route('/objects/<int:object_id>')
... def object_view(object_id):
...     with permission_factory({'id': object_id}).require(http_exception=404):
...         return 'Protected index page'

.. note::

    Invenio source code almost exclusively use the permission factory approach
    for protecting views. In addition usually the permission factory is
    configurable so that Invenio instances can fully override the internal
    permission handling.

Security considerations
~~~~~~~~~~~~~~~~~~~~~~~
We can now test the two views via the built-in Flask test client, and see that
anonymous requests are denied in both cases:

>>> with app.test_client() as c:
...     c.get('/')
<Response streamed [403 FORBIDDEN]>
>>> with app.test_client() as c:
...     c.get('/objects/42')
<Response streamed [404 NOT FOUND]>

In the two above examples for protecting views, you will notice that in one we
return an HTTP 403 Forbidden error, and in the other we return a HTTP 404 Not
Found error.

In views, you should always make a conscious decision if you should return
401/403 or 404 errors as it has important security considerations.

- **403/401** errors exposes *existence* of an object under a given URL. Hence,
  by using 401/403 errors, the system is "leaking" knowledge that certain
  objects exists in the system. Hence, only use 401/403 errors when this
  behavior is desired. In all other cases use 404 errors.
- **404** errors does not leak any additional information and
  *should be the default error* used when a permission check fails.

Superuser
---------
Invenio-Access provides a way to grant superuser privileges to users or roles
via a superuser action. Granting superuser access to a user implicitly gives
that user access to any action in the system without explicitly having to grant
the action.

For instance, currently Bob does not have permissions on our
``view_index_action``:

>>> permission = Permission(view_index_action)
>>> permission.allows(bob_identity)
False

We can however grant Bob superuser access like this:

>>> from invenio_access.permissions import superuser_access
>>> db.session.add(ActionUsers.allow(superuser_access, user=bob))
>>> db.session.commit()

Now, Bob will have access to the ``view_index_action`` even though we did not
explicitly grant Bob access:

>>> permission.allows(bob_identity)
True

System roles
------------
Invenio-Access, in addition to roles defined by the administrator, provides
also system roles. System roles are defined by the system and automatically
assigned to users.

By default the following system roles exists:

* Any user (guests and autenticated users)
* Authenticated user

System roles works very much like normal roles, so you can e.g. assign actions
to them:

>>> from invenio_access import ActionSystemRoles, any_user
>>> db.session.add(ActionSystemRoles.allow(
...     view_index_action, role=any_user))

In order to test system roles from the shell, we have to manually add the need
into the identity.

>>> anonymous.provides.add(any_user)

Now we can check the permission:

>>> permission = Permission(view_index_action)
>>> permission.allows(anonymous)
True

Creating system roles
~~~~~~~~~~~~~~~~~~~~~
Invenio modules may provide additional system roles. You could, for instance
create a system role that could be used to grant permissions based on IP
address.

First the module should define the system role:

>>> from invenio_access import SystemRoleNeed
>>> campus_user = SystemRoleNeed('campus_user')

Next, connect a receiver to the :py:data:`~identity_loaded`
signal and add the system role need to the identity:

>>> from flask import request
>>> from flask_principal import identity_loaded
>>> @identity_loaded.connect_via(app)
... def on_identity_loaded(sender, identity):
...     if request.remote_addr.startswith('192.168.'):
...         identity.provides.add(campus_user)

Last, you need to register the system role in the Invenio module's entry
points in ``setup.py``:

.. code-block:: python

    entry_points={
        'invenio_access.system_roles': [
            'campus_user'
            ' = mymodule.permissions:campus_user',
        ],
    }

Registering actions
-------------------
All actions that a package provides should be registered in the
``invenio_access.actions`` entry points. This ensures the actions are e.g.
available in the adminstration interface and via the CLI.

Below is an example for of the entry point part of the ``setup.py``:

.. code-block:: python

    entry_points={
        'invenio_access.actions': [
            # Action with parameter
            'object_read_action_all'
            ' = mymodule.permissions:object_read_action_all',
            # Action without parameter
            'view_index_action'
            ' = mymodule.permissions:view_index_action',
        ],
    }

Note that for action with parameters you need point to the import path of the
action representing *any* parameters.

Listing actions
---------------
In order to discover which actions are available in a given installation, one
can retrieve them via:

>>> sorted(app.extensions['invenio-access'].actions.keys())
['superuser-access']
"""

from __future__ import absolute_import, print_function

from flask_principal import ActionNeed

from .ext import InvenioAccess
from .factory import action_factory
from .models import ActionRoles, ActionSystemRoles, ActionUsers
from .permissions import ParameterizedActionNeed, Permission, SystemRoleNeed, \
    any_user, authenticated_user, superuser_access
from .proxies import current_access
from .version import __version__

__all__ = (
    '__version__',
    'any_user',
    'authenticated_user',
    'action_factory',
    'current_access',
    'ActionNeed',
    'InvenioAccess',
    'ParameterizedActionNeed',
    'Permission',
    'superuser_access'
    'SystemRoleNeed',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test for admin view."""

from flask import current_app
from invenio_accounts import InvenioAccounts
from werkzeug.local import LocalProxy

from invenio_access.admin import action_roles_adminview, \
    action_system_roles_adminview, action_users_adminview

_datastore = LocalProxy(
    lambda: current_app.extensions['security'].datastore
)


def test_admin(app):
    """Test flask-admin interace."""
    assert isinstance(action_roles_adminview, dict)
    assert isinstance(action_users_adminview, dict)
    assert isinstance(action_system_roles_adminview, dict)

    assert 'model' in action_roles_adminview
    assert 'modelview' in action_roles_adminview
    assert 'model' in action_users_adminview
    assert 'modelview' in action_users_adminview
    assert 'model' in action_system_roles_adminview
    assert 'modelview' in action_system_roles_adminview



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Identity changed loaders."""

from flask_security import current_user

from .permissions import any_user, authenticated_user


def load_permissions_on_identity_loaded(sender, identity):
    """Add system roles "Needs" to users' identities.

    Every user gets the **any_user** Need.
    Authenticated users get in addition the **authenticated_user** Need.
    """
    identity.provides.add(
        any_user
    )
    # if the user is not anonymous
    if current_user.is_authenticated:
        # Add the need provided to authenticated users
        identity.provides.add(
            authenticated_user
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Prevent NULL in ActionUsers for stability reasons."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'dbe499bfda43'
down_revision = '2069a982633b'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.drop_constraint(u'fk_access_actionsusers_user_id_accounts_user',
                       'access_actionsusers', type_='foreignkey')
    op.drop_index(op.f('ix_access_actionsusers_user_id'),
                  table_name='access_actionsusers')

    op.alter_column('access_actionsusers', 'user_id', nullable=False,
                    existing_type=sa.Integer())

    op.create_index(op.f('ix_access_actionsusers_user_id'),
                    'access_actionsusers', ['user_id'], unique=False)
    op.create_foreign_key(op.f('fk_access_actionsusers_user_id_accounts_user'),
                          'access_actionsusers', 'accounts_user', ['user_id'],
                          ['id'], ondelete='CASCADE')


def downgrade():
    """Downgrade database."""
    op.drop_constraint(u'fk_access_actionsusers_user_id_accounts_user',
                       'access_actionsusers', type_='foreignkey')
    op.drop_index(op.f('ix_access_actionsusers_user_id'),
                  table_name='access_actionsusers')

    op.alter_column('access_actionsusers', 'user_id', nullable=True,
                    existing_type=sa.Integer())

    op.create_index(op.f('ix_access_actionsusers_user_id'),
                    'access_actionsusers', ['user_id'], unique=False)
    op.create_foreign_key(op.f('fk_access_actionsusers_user_id_accounts_user'),
                          'access_actionsusers', 'accounts_user', ['user_id'],
                          ['id'], ondelete='CASCADE')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Indentity loaders."""

from flask import current_app, g
from flask_principal import AnonymousIdentity, UserNeed, identity_changed, \
    identity_loaded
from flask_security.utils import login_user, logout_user
from invenio_accounts import testutils

from invenio_access import InvenioAccess
from invenio_access.loaders import load_permissions_on_identity_loaded
from invenio_access.permissions import any_user, authenticated_user


def test_load_permissions_on_request_loaded(app):
    """Checks that the needs are loaded during request in the user Identity."""
    InvenioAccess(app)
    with app.test_request_context():
        with app.test_client() as client:
            client.get('/')
            assert g.identity.provides == {any_user}


def test_load_permissions_on_identity_loaded(app):
    """Check that the needs are loaded properly in the user Identity."""
    InvenioAccess(app)

    with app.test_request_context():
        identity_changed.send(current_app._get_current_object(),
                              identity=AnonymousIdentity())
        assert g.identity.provides == {any_user}

    with app.test_request_context():
        user = testutils.create_test_user('test@example.org')
        login_user(user)
        assert g.identity.provides == {
            any_user, authenticated_user, UserNeed(user.id)
        }
        logout_user()
        # FIXME: The user is still authenticatd when the identity loader
        # is called during logout. We could filter on AnonymousIdentity, but
        # This would be inconsistent as the UserNeed(user.id) is still there.
        # This will pass even if it is unexpected:
        # assert g.identity.provides == {
        #     any_user, authenticated_user, UserNeed(user.id)
        # }
        # Forcing the identity to reload again cleans the mess. In practice
        # this won't be needed as the identity is reloaded between requests.
        identity_changed.send(current_app._get_current_object(),
                              identity=AnonymousIdentity())
        assert g.identity.provides == {any_user}


def test_disabled_loader(app):
    """Check that system role loading function can be disabled."""
    app.config.update(ACCESS_LOAD_SYSTEM_ROLE_NEEDS=False)
    InvenioAccess(app)

    assert load_permissions_on_identity_loaded not in \
        identity_loaded.receivers.values()

    with app.test_request_context():
        identity_changed.send(current_app._get_current_object(),
                              identity=AnonymousIdentity())
        assert g.identity.provides == set()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Utility functions for Invenio-Access."""

from __future__ import absolute_import, print_function

from flask_principal import Identity, RoleNeed, UserNeed


def get_identity(user):
    """Create an identity for a given user instance.

    Primarily useful for testing.
    """
    identity = Identity(user.id)

    if hasattr(user, 'id'):
        identity.provides.add(UserNeed(user.id))

    for role in getattr(user, 'roles', []):
        identity.provides.add(RoleNeed(role.name))

    identity.user = user
    return identity



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create access tables."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '2f63be7b7572'
down_revision = '67ba0de65fbb'
branch_labels = ()
depends_on = '9848d0149abd'


def upgrade():
    """Upgrade database."""
    op.create_table(
        'access_actionsroles',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('action', sa.String(length=80), nullable=True),
        sa.Column('exclude', sa.Boolean(name='exclude'), server_default='0',
                  nullable=False),
        sa.Column('argument', sa.String(length=255), nullable=True),
        sa.Column('role_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['role_id'], [u'accounts_role.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('action', 'exclude', 'argument', 'role_id',
                            name='access_actionsroles_unique')
    )
    op.create_index(
        op.f('ix_access_actionsroles_action'), 'access_actionsroles',
        ['action'], unique=False)
    op.create_index(
        op.f('ix_access_actionsroles_argument'), 'access_actionsroles',
        ['argument'], unique=False)
    op.create_table(
        'access_actionsusers',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('action', sa.String(length=80), nullable=True),
        sa.Column('exclude', sa.Boolean(name='exclude'), server_default='0',
                  nullable=False),
        sa.Column('argument', sa.String(length=255), nullable=True),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['user_id'], [u'accounts_user.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('action', 'exclude', 'argument', 'user_id',
                            name='access_actionsusers_unique')
    )
    op.create_index(
        op.f('ix_access_actionsusers_action'), 'access_actionsusers',
        ['action'], unique=False)
    op.create_index(
        op.f('ix_access_actionsusers_argument'), 'access_actionsusers',
        ['argument'], unique=False)


def downgrade():
    """Downgrade database."""
    op.drop_index(
        op.f('ix_access_actionsusers_argument'),
        table_name='access_actionsusers'
    )
    op.drop_index(
        op.f('ix_access_actionsusers_action'),
        table_name='access_actionsusers'
    )
    op.drop_table('access_actionsusers')
    op.drop_index(
        op.f('ix_access_actionsroles_argument'),
        table_name='access_actionsroles'
    )
    op.drop_index(
        op.f('ix_access_actionsroles_action'),
        table_name='access_actionsroles'
    )
    op.drop_table('access_actionsroles')
